#include "cfx.h"
#include "../include/cfx-internal.h"

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/time.h>


#ifdef WIN32

/* GCC doesn't parse CR as in CR/NL end of lines
   It isn't always possible to insure we don't get CRs,
   we need to have a filter.
 */


#undef getc

int 
getc(FILE *fp)
{
    int c = fgetc(fp);

    return (c == '\r') ? fgetc(fp) : c;
}

#endif


#define FILE_HEADER             "# File initialized at %s\n"
#define CPP_HEADER              "# This written by cpp [%s]\n"
#define CC1_HEADER              "# This written by cc1 [%s]\n"

#define CFX_FILENAME_PREFIX     ".cfx."
#define CPP_PREFIX              "cpp."
#define CC1_PREFIX              "cc1."
#define IFC_PREFIX              ".ifc."
#define CFX_DEFAULT_CPP_FILE    ".cfx.cpp"
#define CFX_DEFAULT_CC1_FILE    ".cfx.cc1"

#define CFX_MIN_INFO_FILE       "cfx.min"
#define MIN_INFO_FILE_SUFFIX    ".cfx.min"
#define CFX_MR_INFO_FILE        "cfx.mr"
#define MR_INFO_FILE_SUFFIX     ".cfx.mr"
#define CFX_HR_INFO_FILE        "cfx.hr"
#define HR_INFO_FILE_SUFFIX     ".cfx.hr"

#define UNDEFINED_ITEM          "*** Undefined Item ***"

#define FCN                     0
#define VAR                     1

#define BUFFER_SIZE             2048

#define ITEM_PTR_RECS_EQUAL(p1,p2)      \
               ((p1->objectLine == p2->objectLine) && \
                (p1->subjectLine == p2->subjectLine) && \
                (p1->objectFile == p2->objectFile) && \
                (p1->subjectFile == p2->subjectFile) && \
                (p1->fileAttributed == p2->fileAttributed) && \
                (p1->kind == p2->kind) && \
                (strcmp (p1->scopeName, p2->scopeName) == 0) && \
                (strcmp (p1->itemName, p2->itemName) == 0))

enum {NONE, CPP, CC1, IFC, HR, MR};

/* Cfx Options */
int     cfx_oneFileOption = 0;
int     cfx_allFactsOption = 0;
int     cfx_noMrOption = 0;
int     cfx_noHrOption = 0;
int     cfx_noMinOption = 0;
int     cfx_noIfcOption = 0;
int     cfx_directOption = 0;

/* Variables used in cfx_cpp and cfx_cc1 */
int     cfx_mode = 0;
int     cfx_allFacts = 0;
char    cfx_macroStartMarker [] = " __MS__";
char    cfx_macroEndMarker [] = " __ME__ ";
char    cfx_macroStartToken [] = "__MS__";
char    cfx_macroEndToken [] = "__ME__";

/* Variables used by cfx_cpp */
int     cfx_suppressMacroMarker = 0;

/* Variables used by cfx_cc1 */
char    *cfx_baseName = NULL;
char    *cfx_cppName = NULL;
int     cfx_macroStartTokenLen = 6;

/* Variables used by cfx_cc */
int     cfx_externalFileNumber;

static FILE     *interface = NULL;
static FILE     *srcFile = NULL;

typedef struct File
{
    char        *fileName;
    char        *shortFileName;
    int         mainFile;
    struct File *next;
} File;

typedef struct Files
{
    char        *fileName;
    char        *shortFileName;
    int         comparisonValue;
    int         mainFile;
    int         interfaceFileWritten;
} Files;

typedef struct MacroStack
{
    int                 fileNumber;
    struct MacroStack   *next;
} MacroStack;

/* Static functions */
static void     AddToInterface ();
static int      CheckReadable ();
static int      CheckWriteable ();
static int      CompareItems ();
static Item     **ConvertDataLLToTableAndSort ();
static void     ConvertFileLLToTable ();
static void     CorrelateLinkRefsAndDefs ();
static void     CorrelateRefsDefsAndDecls ();
static int      CountElementsInDataLL ();
static Item     *CreateNewItem ();
static void     FileReadError ();
static char     *GetFileNameForRead ();
static char     *GetFileNameForWrite ();
static void     InitializeFile ();
static void     MallocString ();
static void     MallocStringLen ();
static void     ReadAndVerifyHeader ();
static void     ReadFileListIntoTable ();
static void     ReadMRDataIntoDataLL ();
static void     SkipAn ();
static void     SkipComments ();
static void     Cfx_AssertFailed ();
static int      SortByName ();
static char     *TimeString ();
static void     WriteFileListFromFileLL ();
static void     WriteFileListFromFileTable ();
static int      WriteHRDataFromDataTable ();
static void     WriteHRExport ();
static void     WriteHRImport ();
static void     WriteMRDataFromDataLL ();
static void     WriteMRDataFromDataLL_gerry ();   /*GK*/
static int      WriteMRDataFromDataTable ();

/* Static variables */
static int      noMrOption, noHrOption, noMinOption;    /* options to cfx_cc */
static int      allFactsOption, oneFileOption;          /* options to cfx_cc */
static int      dirAttribOption;                        /* options to cfx_cc */
static int      noIfcOption;                            /* options to cfx_cc */
static Item     *mainHead = NULL;
static Item     *masterHead = NULL;

static MacroStack       *stackHead = NULL;

static File     *fileHead = NULL;

static char     *emptyString = "";
static char     *cppInfoFileName;
static char     *cc1InfoFileName;

static int      numFiles;
static Files    *fileTable;

static int      currentFileNumber = -1;

static char     *mrInfoFile, *hrInfoFile, *minInfoFile;


#ifdef CFX_CC1
/************************************************************************/
/* These routines are called by "cc1"                                   */
/************************************************************************/
/************************************************************************/
/*      void Cfx_Cc1Initialize (int allFacts, directAttribution,        */
/*                              int noIfc,                              */
/*                              char *baseFileName, char *inFileName,   */
/*                              char *outFileName, char *origFileName)  */
/*                                                                      */
/*   cfx_cc1 {<inDir>/<inFile>} {-o <outDir>/<outFile>} \               */
/*           {-sbasename <bsDir>/<bsName>}                              */

/*   This routine first finds the name of the cpp info file.  There are */
/* several possibilities, so we have to try each one:                   */
/*                                                                      */
/*   inFileName = <inDir>/<inFile> and baseFileName = <bsDir>/<bsFile>  */
/* 1st choice of CFX cpp info file is <inDir>/.cfx.cpp.<inFile>         */
/* 2nd choice of CFX cpp info file is <bsDir>/.cfx.cpp.<bsFile>         */
/* 3rd choice of CFX cpp info file is .cfx.cpp.<bsFile>                 */
/* 4th choice of CFX cpp info file is .cfx.cpp                          */
/*                                                                      */
/*   This routine also determines the name of the output file to be     */
/* for the CFX cc1 information.  It also sets all the options used by   */
/* the CFX cc1 routines.                                                */
/*                                                                      */
/*   inFileName = <outDir>/<outFile> and baseFileName = <bsDir>/<bsFile>*/
/* 1st choice of CFX cc1 info file is <outDir>/.cfx.cc1.<outFile>       */
/* 2nd choice of CFX cc1 info file is <bsDir>/.cfx.cc1.<bsFile>         */
/* 3rd choice of CFX cc1 info file is .cfx.cc1.<bsFile>                 */
/* 4th choice of CFX cc1 info file is .cfx.cc1                          */
/*                                                                      */
/*   If baseFileName is not assigned, then we use origFileName, which   */
/* should be the same.                                                  */
/************************************************************************/
void Cfx_Cc1Initialize (allFacts, directAttribution, noIfc, baseFileName, 
                        inFileName, outFileName, origFileName)
    int allFacts, directAttribution, noIfc;
    char *baseFileName, *inFileName, *outFileName, *origFileName;
{
    FILE        *f;

    if (baseFileName == NULL)
    {
        baseFileName = origFileName;
    }

    cppInfoFileName = GetFileNameForRead (baseFileName, inFileName, CPP);

/* The following line was commented out by NB */
/*  cc1InfoFileName = GetFileNameForWrite (baseFileName, outFileName, CC1); */

/* The following lines were added by NB */
    if (!CheckWriteable (outFileName))
    {
      fprintf (stderr, "[25] Unable to open \"%s\" for write.\n", outFileName);
      exit (125);
    }
    cc1InfoFileName = (char *) malloc (strlen (outFileName) + 1);
    strcpy(cc1InfoFileName, outFileName);
/* The preceding lines were added by NB */

    f = fopen (cppInfoFileName, RMODE);
    
    if (f == NULL)
    {
        fprintf (stderr, 
                 "[4] Cfx cpp info file \"%s\" could not be found by cc1!\n", 
                 cppInfoFileName);
        exit (104);
    }

    /*******************/
    /* Read CPP header */
    /*******************/
    ReadAndVerifyHeader (f, cppInfoFileName, CPP_HEADER);

    /* Read the file list into the file table */
    ReadFileListIntoTable (f);
   
    /* Read File Data */
    ReadMRDataIntoDataLL (f);

    fclose (f);

    allFactsOption = allFacts;
    dirAttribOption = directAttribution;
    noIfcOption = noIfc;
} /* Cfx_Cc1Initialize */


/************************************************************************/
/*      void Cfx_Cc1Finalize ()                                         */
/************************************************************************/
void Cfx_Cc1Finalize ()
{
    FILE        *f;

    /* Collect, sort and output Include Export items */
    f = fopen (cc1InfoFileName, WMODE);                   /* Open File */
    fprintf (f, CC1_HEADER, TimeString ());             /* Write header */
    WriteFileListFromFileTable (f);                     /* Write file list */
    CorrelateRefsDefsAndDecls (mainHead);
    WriteMRDataFromDataLL_gerry (f, mainHead);                /* Write MR data */
    fclose (f);                                         /* Close file */
    if (interface == NULL)
    {
        fclose (interface);
        fclose (srcFile);
    }
} /* Cfx_Cc1Finalize */


/************************************************************************/
/*      int     Cfx_GetFileNumber (immediate)                           */
/************************************************************************/
int     Cfx_GetFileNumber (immediate, fileName)
    int         immediate;
    char        *fileName;
{
    static int  lastNumber = -599;
    static char lastName [4096] = "\377";
    int         cnt;

    /* fprintf(stderr, "GetFileNum: '%s'\n", fileName); */

    if (fileName == NULL) return (NULL_FILENAME_FILENUM);

    fileName = (char *) Cfx_CanonPath(fileName);                       /*NB*/

    if (!dirAttribOption && !immediate && (stackHead != NULL))
    {
        return (stackHead->fileNumber);
    }

    if (strcmp (fileName, lastName) == 0)
    {
        return (lastNumber);
    }

    for (cnt = 1 ; cnt <= numFiles ; cnt++)
    {
        if (strcmp (fileName, fileTable [cnt].fileName) == 0) 
        {
            lastNumber = cnt;
            strcpy (lastName, fileName);
            return (cnt);
        }
    }

    fprintf (stderr, "[5] Unable to find file name to match \"%s\",\n", 
             fileName);
    exit (105);
} /* Cfx_GetFileNumber */


/************************************************************************/
/*      void    Cfx_PushMacroStack (char fileNumber)                    */
/************************************************************************/
void    Cfx_PushMacroStack (fileNumber)
    char        *fileNumber;
{
    MacroStack  *newMacroStack = (MacroStack *) malloc (sizeof (MacroStack));
    char        *ptr = fileNumber;

    if (!dirAttribOption)
    {
        if (newMacroStack == NULL)
        {
            fprintf (stderr, "[6] No memory to allocate macro stack\n");
            exit (106);
        }

        newMacroStack->fileNumber = 0;
        newMacroStack->next = stackHead;

        while (*ptr != 0)
        {
            newMacroStack->fileNumber *= 10;
            newMacroStack->fileNumber += (*ptr - '0');
            ptr++;
        }

        if ((newMacroStack->fileNumber < 1) || 
            (newMacroStack->fileNumber > 10000))
        {
            fprintf (stderr, 
                 "[7] Bad file number of \"%s\" passed to Cfx_PushMacroStack\n",
                     fileNumber);
            exit (107);
        }
            
        stackHead = newMacroStack;
    }
} /* Cfx_PushMacroStack */


/************************************************************************/
/*      void    Cfx_PopMacroStack (void)                                */
/************************************************************************/
void    Cfx_PopMacroStack ()
{
    if (!dirAttribOption)
    {
        if (stackHead == NULL)
        {
            fprintf (stderr, 
                     "[8] Failed in attempting to pop empty macro stack\n");
            exit (108);
        }
        stackHead = stackHead->next;
    }
} /* Cfx_PopMacroStack */


/************************************************************************/
/*      void Cfx_EnumItemDefinition (int fileAttributed,                */
/*                                    int fileDefiningItem,             */
/*                                    int lineDefiningItem,             */
/*                                    char *itemName)                   */
/************************************************************************/
void Cfx_EnumItemDefinition (fileAttributed, fileDefiningItem, 
                              lineDefiningItem, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                             lineDefiningItem, EXPORT_ENUM_ITEM,UNDEFINED_ITEM);
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_EnumItemDefinition */


/************************************************************************/
/*      void Cfx_EnumTagDefinition (int fileAttributed,                 */
/*                                   int fileDefiningItem,              */
/*                                   int lineDefiningItem,              */
/*                                   char *itemName)                    */
/************************************************************************/
void Cfx_EnumTagDefinition (fileAttributed, fileDefiningItem, 
                             lineDefiningItem, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                             lineDefiningItem, EXPORT_ENUM_TAG,UNDEFINED_ITEM);
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_EnumTagDefinition */


/************************************************************************/
/*      void Cfx_FcnDeclaration (int fileAttributed,                    */
/*                                int fileDefiningItem,                 */
/*                                int lineDefiningItem,                 */
/*                                char *itemName)                       */
/************************************************************************/
void Cfx_FcnDeclaration (fileAttributed, fileDefiningItem, 
                          lineDefiningItem, public, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    int         public;
    char        *itemName;
{
    Item        *newItem;

    if (public)
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_PUBLIC_FCN_DECL,UNDEFINED_ITEM);
    }
    else
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_STATIC_FCN_DECL, UNDEFINED_ITEM);
    }
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_FcnDeclaration */


/************************************************************************/
/*      void Cfx_GlobalFcnDefinition (int fileAttributed,               */
/*                                     int fileDefiningItem,            */
/*                                     int lineDefiningItem,            */
/*                                     int public,                      */
/*                                     char *itemName)                  */
/************************************************************************/
CFX_Item Cfx_GlobalFcnDefinition (fileAttributed, fileDefiningItem, 
                               lineDefiningItem, public, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    int         public;
    char        *itemName;
{
    Item        *newItem;

    if (public)
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_PUBLIC_FCN_DEF, UNDEFINED_ITEM);
        if (!noIfcOption)
        {
            AddToInterface (itemName, fileDefiningItem, lineDefiningItem, FCN);
        }
    }
    else
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_STATIC_FCN_DEF, UNDEFINED_ITEM);
    }
    MallocString (&(newItem->itemName), itemName);

    return newItem;
} /* Cfx_GlobalFcnDefinition */


/************************************************************************/
/*      void Cfx_GlobalVarDefinition (int fileAttributed,               */
/*                                     int fileDefiningItem,            */
/*                                     int lineDefiningItem,            */
/*                                     int public,                      */
/*                                     char *itemName)                  */
/************************************************************************/
void Cfx_GlobalVarDefinition (fileAttributed, fileDefiningItem, 
                               lineDefiningItem, public, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    int         public;
    char        *itemName;
{
    Item        *newItem;

    if (public)
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_PUBLIC_VAR_DEF, UNDEFINED_ITEM);
        if (!noIfcOption)
        {
            AddToInterface (itemName, fileDefiningItem, lineDefiningItem, VAR);
        }
    }
    else
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_STATIC_VAR_DEF, UNDEFINED_ITEM);
    }
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_GlobalVarDefinition */


/************************************************************************/
/*      void Cfx_StructOrUnionDefinition (int fileAttributed,           */
/*                                         int fileDefiningItem,        */
/*                                         int lineDefiningItem,        */
/*                                         int isStruct,                */
/*                                         char *itemName)              */
/************************************************************************/
void Cfx_StructOrUnionDefinition (fileAttributed, fileDefiningItem, 
                                   lineDefiningItem, isStruct, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    int         isStruct;
    char        *itemName;
{
    Item        *newItem;

    if (isStruct)
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_STRUCT, UNDEFINED_ITEM);
    }
    else
    {
        newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                                 lineDefiningItem, EXPORT_UNION, UNDEFINED_ITEM);
    }
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_StructOrUnionDefinition */


/************************************************************************/
/*      void Cfx_TypeDefinition (int fileAttributed,                    */
/*                                int fileDefiningItem,                 */
/*                                int lineDefiningItem,                 */
/*                                char *itemName)                       */
/************************************************************************/
void Cfx_TypeDefinition (fileAttributed, fileDefiningItem, 
                          lineDefiningItem, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                             lineDefiningItem, EXPORT_TYPE, UNDEFINED_ITEM);
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_TypeDefinition */


/************************************************************************/
/*      void Cfx_VarDeclaration (int fileAttributed,                    */
/*                                int fileDefiningItem,                 */
/*                                int lineDefiningItem,                 */
/*                                char *itemName)                       */
/************************************************************************/
void Cfx_VarDeclaration (fileAttributed, fileDefiningItem, 
                          lineDefiningItem, itemName)
    int         fileAttributed;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileDefiningItem, 
                             lineDefiningItem, EXPORT_PUBLIC_VAR_DECL, UNDEFINED_ITEM);
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_VarDeclaration */


/***********************************************************************//*SS*/
/*      void Cfx_EnumTagDefinitionEnd (int fileAttributed,             *//*SS*/
/*                                int fileDefiningItem,                *//*SS*/
/*                                int endLineDefiningItem,             *//*SS*/
/*                                char *itemName)                      *//*SS*/
/***********************************************************************//*SS*/

extern void Cfx_EnumTagDefinitionEnd (fileAttributed, fileDefiningItem,
					endLineDefiningItem, itemName)				/*SS*/
	int		fileAttributed;                                     /*SS*/
    int		fileDefiningItem;                                   /*SS*/
    int		endLineDefiningItem;                                /*SS*/
    char	*itemName;                                          /*SS*/
{                                                               /*SS*/
	Item	*newItem;                                           /*SS*/
                                                                /*SS*/
	newItem = CreateNewItem(fileAttributed, fileDefiningItem,   /*SS*/
				endLineDefiningItem, EXPORT_ENUM_TAG_END,       /*SS*/
                UNDEFINED_ITEM);                                /*SS*/
                                                                /*SS*/
	MallocString (&(newItem->itemName), itemName);              /*SS*/
} /* Cfx_EnumTagDeclarationEnd */								/*SS*/

/**********************************************************************/ /*SS*/
/*    void Cfx_StructOrUnionDefinitionEnd (int fileAttributed,        */ /*SS*/ 
/*                                       int fileDefiningItem,        */ /*SS*/
/*                                       int endLineDefiningItem,     */ /*SS*/
/*                                       int isStruct,                */ /*SS*/
/*                                       char *itemName)              */ /*SS*/
/**********************************************************************/ /*SS*/
void Cfx_StructOrUnionDefinitionEnd (fileAttributed, fileDefiningItem,   /*SS*/
                                   endLineDefiningItem, isStruct, itemName)
    int         fileAttributed;                                   /*SS*/
    int         fileDefiningItem;                                 /*SS*/
    int         endLineDefiningItem;                              /*SS*/
    int         isStruct;                                         /*SS*/
    char        *itemName;                                        /*SS*/
{                                                                 /*SS*/
    Item        *newItem;                                         /*SS*/
                                                                  /*SS*/
    if (isStruct)                                                 /*SS*/
    {                                                             /*SS*/
        newItem = CreateNewItem (fileAttributed, fileDefiningItem,/*SS*/
                        endLineDefiningItem, EXPORT_STRUCT_END,   /*SS*/
                        UNDEFINED_ITEM);                          /*SS*/
    }                                                             /*SS*/
    else                                                          /*SS*/
    {                                                             /*SS*/
        newItem = CreateNewItem (fileAttributed, fileDefiningItem,/*SS*/
                        endLineDefiningItem, EXPORT_UNION_END,    /*SS*/
                        UNDEFINED_ITEM);                          /*SS*/
    }                                                             /*SS*/
    MallocString (&(newItem->itemName), itemName);                /*SS*/
} /* Cfx_StructOrUnionDefinitionEnd */

#if 0
/**********************************************************************/ /*SS*/
/*      void Cfx_GlobalFcnDefinitionEnd (int fileAttributed,          */ /*SS*/
/*                                     int fileDefiningItem,          */ /*SS*/
/*                                     int lineDefiningItem,          */ /*SS*/
/*                                     int public,                    */ /*SS*/
/*                                     char *itemName)                */ /*SS*/
/**********************************************************************/ /*SS*/
void Cfx_GlobalFcnDefinitionEnd (fileAttributed, fileDefiningItem,       /*SS*/
                               endLineDefiningItem, public, itemName)    /*SS*/
    int         fileAttributed;                                          /*SS*/
    int         fileDefiningItem;                                        /*SS*/
    int         endLineDefiningItem;                                     /*SS*/
    int         public;                                                  /*SS*/
    char        *itemName;                                               /*SS*/
{                                                                        /*SS*/
    Item        *newItem;                                                /*SS*/
                                                                         /*SS*/
    if (public)                                                          /*SS*/
    {                                                                    /*SS*/
        newItem = CreateNewItem (fileAttributed, fileDefiningItem,       /*SS*/
                        endLineDefiningItem, EXPORT_PUBLIC_FCN_DEF_END,  /*SS*/
                        UNDEFINED_ITEM);                                 /*SS*/
    }                                                                    /*SS*/
    else                                                                 /*SS*/
    {                                                                    /*SS*/
        newItem = CreateNewItem (fileAttributed, fileDefiningItem,       /*SS*/
                        endLineDefiningItem, EXPORT_STATIC_FCN_DEF_END,  /*SS*/
                        UNDEFINED_ITEM);                                 /*SS*/
    }                                                                    /*SS*/
    MallocString (&(newItem->itemName), itemName);                       /*SS*/
} /* Cfx_GlobalFcnDefinitionEnd */                                       /*SS*/

#else

/**********************************************************************/ /*GF*/
/*      void Cfx_GlobalFcnDefinitionEnd (Item *item,                  */ /*GF*/
/*                                     int lineDefiningItem,          */ /*GF*/
/*                                     int public)                    */ /*GF*/
/**********************************************************************/ /*GF*/
void Cfx_GlobalFcnDefinitionEnd (item, endLineDefiningItem, public)      /*GF*/
    Item        *item;							 /*GF*/
    int         endLineDefiningItem;                                     /*GF*/
    int         public;                                                  /*GF*/
{                                                                        /*GF*/
    Item        *newItem;                                                /*GF*/

    if (public)                                                          /*GF*/
    {                                                                    /*GF*/
        newItem = CreateNewItem (item->fileAttributed,			 /*GF*/ 
				 item->subjectFile,		 	 /*GF*/
                                 endLineDefiningItem,			 /*GF*/ 
			         EXPORT_PUBLIC_FCN_DEF_END,              /*GF*/
                        	 UNDEFINED_ITEM);                        /*GF*/
    }                                                                    /*GF*/
    else                                                                 /*GF*/
    {                                                                    /*GF*/
        newItem = CreateNewItem (item->fileAttributed,			 /*GF*/ 
				 item->subjectFile,			 /*GF*/
                                 endLineDefiningItem,			 /*GF*/ 
			         EXPORT_STATIC_FCN_DEF_END,              /*GF*/
                        	 UNDEFINED_ITEM);                        /*GF*/
    }                                                                    /*GF*/
    MallocString (&(newItem->itemName), item->itemName);                 /*GF*/
} /* Cfx_GlobalFcnDefinitionEnd */                                       /*GF*/
#endif 

#endif /* CFX_CC1 */

/************************************************************************/
/* These routines are called by more than one cfx program               */
/************************************************************************/
/************************************************************************/
/*      int     Cfx_AddFile (char *fileName, *shortName, int len,       */
/*                           int angle_brackets)                        */
/************************************************************************/
int     Cfx_AddFile (fileName, shortName, len, angle_brackets)
    char *fileName, *shortName;
    int len, angle_brackets;
{
    File        *ptr = fileHead, *lastPtr = NULL, *newFile;
    int         fileNo = 0;

    fileName = (char *) Cfx_CanonPath(fileName);                        /*NB*/

    /* First, check to see if it's already in the list */
    while (ptr != NULL)
    {
        fileNo++;
        if (strcmp (ptr->fileName, fileName) == 0)
        {
            if (len == -1)
            {
                ptr->mainFile = 0;
            }
            /* Found it! */
            return (fileNo);
        }
        lastPtr = ptr;
        ptr = ptr->next;
    }

    /* Okay it's not there.  Add it to the list. */
    newFile = (File *) malloc (sizeof (File));
    if (newFile == NULL)
    {
        fprintf (stderr, "[3] No memory for File \"%s\".\n", fileName);
        exit (103);
    }

    if ((len == -1) || (len == -2))
    {
        newFile->fileName = fileName;
        newFile->shortFileName = shortName;
        if (len == -1)
        {
            newFile->mainFile = 0;
        }
        else
        {
            newFile->mainFile = 1;
        }
    }
    else
    {
        MallocString (&(newFile->fileName), fileName);
        if (shortName == NULL)
        {
            newFile->shortFileName = newFile->fileName;
        }
        else
        {
            newFile->shortFileName = (char *) malloc (len + 3);
            if (angle_brackets)
            {
                newFile->shortFileName [0] = '<';
                strncpy (&newFile->shortFileName [1], shortName, len);
                newFile->shortFileName [len + 1] = 0;
                strcat (newFile->shortFileName, ">");
            }
            else
            {
                strncpy (newFile->shortFileName, shortName, len);
                newFile->shortFileName [len] = 0;
            }
        }
    }
        
    newFile->next = NULL;
    if (lastPtr == NULL)
    {
        fileHead = newFile;
    }
    else
    {
        lastPtr->next = newFile;
    }
    return (fileNo + 1);
} /* Cfx_AddFile */


/************************************************************************/
/*      void    Cfx_CreateFileName (char *fileName, finalFileName,      */
/*                                  int purpose)                        */
/*                                                                      */
/*  Purpose: Can be one of CPP, CC1, HR, MR, IFC, -1                    */
/*                                                                      */
/*  
/************************************************************************/
void    Cfx_CreateFileName (fileName, finalFileName, purpose)
    char        *fileName, *finalFileName;
    int         purpose;
{
    char        *ptr, *ptr1;

    if (strncmp (fileName, CFX_FILENAME_PREFIX, 
                 strlen (CFX_FILENAME_PREFIX)) == 0)
    {
        strcpy (finalFileName, fileName);
        return;
    }

    if (strrchr (fileName, '/') == NULL)
    {
        if (purpose == IFC)
        {
            finalFileName [0] = 0;
        }
        else
        {
            strcpy (finalFileName, CFX_FILENAME_PREFIX);
        }

        if (purpose == CPP) strcat (finalFileName, CPP_PREFIX);
        else if (purpose == CC1) strcat (finalFileName, CC1_PREFIX);
             else if (purpose == IFC) strcat (finalFileName, IFC_PREFIX);

        strcat (finalFileName, fileName);
    }
    else
    {
        strcpy (finalFileName, fileName);
        ptr = strrchr (finalFileName, '/') + 1;
        *ptr = 0;

        if (purpose != IFC)
        {
            strcpy (ptr, CFX_FILENAME_PREFIX);
        }

        if (purpose == CPP) strcat (finalFileName, CPP_PREFIX);
        else if (purpose == CC1) strcat (finalFileName, CC1_PREFIX);
             else if (purpose == IFC) strcat (finalFileName, IFC_PREFIX);

        ptr = strrchr (fileName, '/') + 1;
        strcat (finalFileName, ptr);
    }

    ptr = finalFileName;
} /* Cfx_CreateFileName */


/************************************************************************/
/*      void Cfx_Import (int importType,                                */
/*                        int fileAttributed,                           */
/*                        int fileUsingItem,                            */
/*                        int fileDefiningItem,                         */
/*                        int lineDefiningItem,                         */
/*                        char *itemName)                               */
/************************************************************************/
void Cfx_Import (importType, fileAttributed, fileUsingItem, 
                  fileDefiningItem, lineDefiningItem, itemName)
    int         importType;
    int         fileAttributed;
    int         fileUsingItem;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileUsingItem, -1, importType, UNDEFINED_ITEM);
    newItem->objectFile = fileDefiningItem;
    newItem->objectLine = lineDefiningItem;
    MallocString (&(newItem->itemName), itemName);
} /* Cfx_Import */


/************************************************************************/
/*      void Cfx_Import_2 (int importType,                              */
/*                        int fileAttributed,                           */
/*                        int fileUsingItem,                            */
/*                        int fileDefiningItem,                         */
/*                        int lineDefiningItem,                         */
/*                        char *itemName                                */
/*                        char *scopeName)                              */
/************************************************************************/
void Cfx_Import_2 (importType, fileAttributed, fileUsingItem, 
                  fileDefiningItem, lineDefiningItem, itemName, scopeName )
    int         importType;
    int         fileAttributed;
    int         fileUsingItem;
    int         fileDefiningItem;
    int         lineDefiningItem;
    char        *itemName;
    char        *scopeName;
{
    Item        *newItem;

    newItem = CreateNewItem (fileAttributed, fileUsingItem, -1, importType,scopeName, UNDEFINED_ITEM);
    newItem->objectFile = fileDefiningItem;
    newItem->objectLine = lineDefiningItem;
/*     MallocString (&(newItem->scopeName), scopeName); */
    MallocString (&(newItem->itemName), itemName);


} /* Cfx_Import_2  */




/************************************************************************/
/* Static Functions                                                     */
/************************************************************************/
#ifdef CFX_CC1
/************************************************************************/
/*      static void     AddToInterface (itemName, fileDefiningItem,     */
/*                                      lineDefiningItem, kind)         */
/************************************************************************/
static void     AddToInterface (itemName, fileDefiningItem, lineDefiningItem,
                                kind)
    char        *itemName;
    int         fileDefiningItem;
    int         lineDefiningItem;
    int         kind;
{
    extern FILE *finput;
    int         line, lineToReadSourceFrom;
    int         pos, lastPos, nextPos, startPos, ptr, ch;
    int         foundEnd, commentEnded;
    char        buffer [BUFFER_SIZE];
    static int  lastFileWrittenTo = -1;
    static int  srcLine = 1;
    static int  srcPos = 0;

    if (fileDefiningItem != lastFileWrittenTo)
    {
        if (interface != NULL)
        {
            fclose (interface);
            fclose (srcFile);
        }
        Cfx_CreateFileName (fileTable [fileDefiningItem].fileName, 
                            buffer, IFC);
        if (fileTable [fileDefiningItem].interfaceFileWritten)
        {
            interface = fopen (buffer, AMODE);
            if (interface == NULL)
            {
                lastFileWrittenTo = -1;
                return;
            }
        }
        else
        {
            fileTable [fileDefiningItem].interfaceFileWritten = 1;
            interface = fopen (buffer, WMODE);
            if (interface == NULL)
            {
                lastFileWrittenTo = -1;
                return;
            }
        }

        srcFile = fopen (fileTable [fileDefiningItem].fileName, RMODE);

        lastFileWrittenTo = fileDefiningItem;
    }

    line = lineDefiningItem;

    /* Get the initial position */
    pos = ftell (finput);

    /* Get to either the beginning of the file, */
    /* or the semicolon before the definition.  */
    lastPos = pos - 2;
    foundEnd = 0;
    do
    {
        nextPos = lastPos - 1024;
        if (nextPos < 0) nextPos = 0;
        fseek (finput, nextPos, 0);
        fread (buffer, lastPos - nextPos + 1, 1, finput);
        ptr = lastPos - nextPos;
        while ((ptr >= 0) && (!foundEnd))
        {
            if ((buffer [ptr] == ';') || (buffer [ptr] == '}')) foundEnd = 1;
            else
            {
                if (buffer [ptr] == '\n')
                {
                    line--;
                }
                ptr--;
            }
        }
        lastPos = nextPos;
    } while ((lastPos != 0) && (!foundEnd));

    startPos = lastPos + ptr + 1;
    fseek (finput, startPos, 0);

    /********************************************************************/
    /* The file pointer just after the end of the line before the       */
    /* definition.  We now go through the line and save any comment     */
    /* lines we might happen across.  We then start writing from the    */
    /* first non-comment, non-whitespace until we read either a semi-   */
    /* colon for an open bracket (for a FCN)                            */
    /********************************************************************/
    fprintf (interface, "%s\n", itemName);

    /********************************************************************/
    /* This is the line that we should read from in the source.         */
    /* However, we need to start at the next line so as not read the    */
    /* semicolon or the close bracket                                   */
    /********************************************************************/
    lineToReadSourceFrom = line;
    if (startPos != 0) lineToReadSourceFrom++;

    /* Now read through until we hit non-white space.  Ignoring # directives */
    do
    {
        ch = fgetc (finput);
        if (ch == '#')
        {
            do ch = getc (finput);
            while (ch != '\n');
        }
        if (ch == '\n')
        {
            line++;
        }
    } while ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') ||
             (ch == '\f'));

    /* We've reached the definition.  Place it in the interface file */
    fprintf (interface, "%d\n", line);
    if (kind == FCN)
    {
        int     parenCount = 1;

        do
        {
            fputc (ch, interface);
            ch = fgetc (finput);
        } while (ch != '(');

        fputc (ch, interface);

        do
        {
            ch = fgetc (finput);
            fputc (ch, interface);
            if (ch == '(') parenCount++;
            if (ch == ')') parenCount--;
        } while (parenCount > 0);
    }
    else
    {
        ptr = 0;
        do
        {
            if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') ||
                (ch == '\f'))
            {
                buffer [ptr++] = ch;
                if (ptr > BUFFER_SIZE - 10)
                {
                    buffer [ptr] = 0;
                    fputs (buffer, interface);
                    ptr = 0;
                }
            }
            else
            {
                if (ptr > 0)
                {
                    buffer [ptr] = 0;
                    fputs (buffer, interface);
                    ptr = 0;
                }
                fputc (ch, interface);
            }
            ch = fgetc (finput);
        } while (ch != ';');
    }

    /****************************/
    /* Now fill in the comments */
    /****************************/
    fprintf (interface, "\n!\n");
    if (srcFile != NULL)
    {
        if (srcLine > lineToReadSourceFrom)
        {
            srcLine = 1;
            srcPos = 0;
        }
        line = srcLine;
        fseek (srcFile, srcPos, 0);

        while (line < lineToReadSourceFrom)
        {
            do ch = fgetc (srcFile);
            while ( (ch != '\n') && (ch != EOF) );             /*NB*/
         /*   if (ch == EOF) {                             */       /*NB*/
         /*     warning ("cfx_cc1: Unexpected EOF");    */          /*NB*/
         /*   }  */                                                 /*NB*/
            line++;
        }

        srcLine = line;
        srcPos = ftell (srcFile);

        do
        {
            ch = fgetc (srcFile);
            if (ch == '/')
            {
                ch = fgetc (srcFile);
                if (ch == '*')
                {
                    /* Comment */
                    fprintf (interface, "/*");
                    commentEnded = 0;
                    do
                    {
                        ch = fgetc (srcFile);
                        fputc (ch, interface);
                        if (ch == '*')
                        {
                            ch = fgetc (srcFile);
                            if (ch == '/')
                            {
                                fputc (ch, interface);
                                fputc ('\n', interface);
                                commentEnded = 1;
                            }
                            else ungetc (ch, srcFile);
                        }
                    }
                    while (!commentEnded);
                    ch = ' ';
                }
                else ungetc (ch, srcFile);
            }
            if (ch == '#')
            {
                do ch = fgetc (srcFile);
                while (ch != '\n');
            }
        } while ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') ||
                 (ch == '\f'));
    } /* if (srcFile != NULL) */

    fprintf (interface, "!\n");
                
    /* Move back to initial position */
    fseek (finput, pos, 0);
} /* AddToInterface */
#endif /* CFX_CC1 */


/************************************************************************/
/*      static int      CheckReadable ()                                */
/************************************************************************/
static int      CheckReadable (fileName)
    char *fileName;
{
    FILE        *f;

    f = fopen (fileName, RMODE);
    if (f == NULL)
    {
        return (0);
    }

    fclose(f);

    return (1);
} /* CheckReadable */


/************************************************************************/
/*      static int      CheckWriteable ()                               */
/************************************************************************/
static int      CheckWriteable (fileName)
    char *fileName;
{
    FILE        *f;

    f = fopen (fileName, WMODE);
    if (f != NULL)
    {
	fclose(f);
        if (unlink (fileName) != 0)
        {
            fprintf (stderr, 
        "[12] Having created CFX info file \"%s\",it could not be deleted!\n", 
                     fileName);
            exit (112);
        }
        return (1);
    }
    return (0);
} /* CheckWriteable */


/************************************************************************/
/*      static int CompareItems (first, second)                         */
/************************************************************************/
static int      CompareItems (first, second)
    Item        **first, **second;
{
    int         comp;
    
    comp = (*first)->fileAttributed - (*second)->fileAttributed;
    if (comp != 0) return (comp);

    comp = (*first)->objectFile - (*second)->objectFile;
    if (comp != 0) return (comp);

    comp = (*first)->kind - (*second)->kind;
    if (comp != 0) return (comp);

    comp = strcmp ((*first)->itemName, (*second)->itemName);
    if (comp != 0) return (comp);

    comp = (*first)->subjectFile - (*second)->subjectFile;
    if (comp != 0) return (comp);

    return ((*first)->subjectLine - (*second)->subjectLine);
} /* CompareItems */


/************************************************************************/
/*      static Item     **ConvertDataLLToTableAndSort (                 */
/*                          Item *listHead, int *numElements,           */
/*                          int (*comparisonFunction)())                */
/************************************************************************/
static Item     **ConvertDataLLToTableAndSort (listHead, numElements, 
                                               comparisonFunction)
    Item        *listHead;
    int         *numElements;
    int         (*comparisonFunction)();
{
    int         i, src, dest;
    Item        *ptr = listHead, **base;

    *numElements = 0;
    while (ptr != NULL)
    {
        if (ptr->kind < IMPORT_FCN_REF_TEMP)
        {
            (*numElements)++;
        }
        ptr = ptr->next;
    }

    base = (Item **) malloc ((*numElements + 1) * sizeof (Item *));
    if (base == NULL)
    {
        fprintf (stderr, "[13] No memory to allocate base\n");
        exit (113);
    }

    ptr = listHead;
    i = 0;
    while (ptr != NULL)
    {
        if (ptr->kind < IMPORT_FCN_REF_TEMP)
        {
            base [i] = ptr;
            i++;
        }
        ptr = ptr->next;
    }

    qsort (base, *numElements, sizeof (Item *), comparisonFunction);

    /* Now go through the table and remove non unique elements */
    src = 1;
    dest = 1;
    while (src < *numElements)
    {
        while (ITEM_PTR_RECS_EQUAL (base [src], base [dest - 1]))
        {
            src++;
        }
        if (dest != src)
        {
            base [dest++] = base [src++];
        }
        else
        {
            src++;
            dest++;
        }
    }

    fprintf (stderr, "%d duplicate elements removed\n", *numElements - dest);

    *numElements = dest;
        
    /* Allocate an element at the end with type NO_KIND */
    base [*numElements] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);

    return (base);
} /* ConvertDataLLToTableAndSort */


/************************************************************************/
/*      static File     *ConvertFileLLToTable (void)                    */
/*                                                                      */
/*   This sets global variables numFiles and fileTable.  It also frees  */
/* the file LL while creating the table.                                */
/************************************************************************/
static void     ConvertFileLLToTable ()
{
    int         i;
    File        *ptr = fileHead, *next;

    numFiles = 0;
    while (ptr != NULL)
    {
        numFiles++;
        ptr = ptr->next;
    }

    fileTable = (Files *) malloc ((numFiles + 1) * sizeof (Files));
    if (fileTable == NULL)
    {
        fprintf (stderr, "[14] No memory for file table.\n");
        exit (114);
    }

    ptr = fileHead;
    for (i = 1 ; i <= numFiles ; i++)
    {
        fileTable [i].fileName = ptr->fileName;
        fileTable [i].shortFileName = ptr->shortFileName;
        fileTable [i].mainFile = ptr->mainFile;
        next = ptr->next;
        free (ptr);
        ptr = next;
    }
    fileHead = NULL;

    /************************/
    /* Set up the zero item */
    /************************/
    fileTable [0].fileName = "(No File)";
    fileTable [0].shortFileName = "(No File)";
    fileTable [0].comparisonValue = 0;
    fileTable [0].mainFile = 0;
} /* ConvertFileLLToTable */


/************************************************************************/
/*      static void     CorrelateLinkRefsAndDefs (Item *listHead)       */
/************************************************************************/
static void     CorrelateLinkRefsAndDefs (listHead)
    Item        *listHead;
{
    int         counts [NO_KIND];
    int         numItems;
    Item        *ptr = listHead;
    Item        **refBase, **defBase;
    int         ref = 0, def = 0, refPtr, defPtr;
    char        *name;
    int         kind;

    (void) CountElementsInDataLL (listHead, counts);

    /* Allocate memory for the references */
    numItems = counts [IMPORT_LINK_FCN_TEMP] + 
               counts [IMPORT_LINK_VAR_TEMP] + 1;
    refBase = (Item **) malloc (numItems * sizeof (Item *));
    if (refBase == NULL)
    {
        fprintf (stderr, "[15] No memory to allocate link refBase\n");
        exit (115);
    }

    /* Allocate memory for the definitions */
    numItems = counts [EXPORT_PUBLIC_FCN_DEF] + 
               counts [EXPORT_PUBLIC_VAR_DEF] + 1;
    defBase = (Item **) malloc (numItems * sizeof (Item *));
    if (defBase == NULL)
    {
        fprintf (stderr, "[16] No memory to allocate defBase\n");
        exit (116);
    }

    while (ptr != NULL)
    {
        switch (ptr->kind)
        {
            case IMPORT_LINK_FCN_TEMP:
            case IMPORT_LINK_VAR_TEMP:
                refBase [ref++] = ptr;
                break;
            case EXPORT_PUBLIC_FCN_DEF:
            case EXPORT_PUBLIC_VAR_DEF:
                defBase [def++] = ptr;
                break;
            default:
                break;
        }
        ptr = ptr->next;
    }

    qsort (refBase, ref, sizeof (Item *), SortByName);
    qsort (defBase, def, sizeof (Item *), SortByName);

    /* Allocate an element at the end with type NO_KIND */
    refBase [ref] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);
    refBase [ref]->itemName = "\377";
    defBase [def] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);
    defBase [def]->itemName = "\377";

    /* First we add the IMPORT_LINK_FCN and IMPORT_LINK_VAR.  For every */
    /* reference, we emit an IMPORT_LINK_FCN/VAR for every definition   */
    /* in a different file                                              */
    defPtr = 0;
    for (refPtr = 0 ; refPtr < ref ; refPtr++)
    {
        name = refBase [refPtr]->itemName;
        while (strcmp (name, defBase [defPtr]->itemName) > 0)
        {
            defPtr++;
        }
        if (strcmp (name, defBase [defPtr]->itemName) == 0)
        {
            if (refBase [refPtr]->fileAttributed != 
                defBase [defPtr]->fileAttributed)
            {
                if (refBase [refPtr]->kind == IMPORT_LINK_FCN_TEMP)
                    kind = IMPORT_LINK_FCN;
                else if (refBase [refPtr]->kind == IMPORT_LINK_VAR_TEMP)
                    kind = IMPORT_LINK_VAR;
                else Cfx_AssertFailed (4);
                ptr = CreateNewItem (refBase [refPtr]->fileAttributed,
                                     refBase [refPtr]->subjectFile,
                                     refBase [refPtr]->subjectLine,
                                     NO_KIND, UNDEFINED_ITEM);
                ptr->kind = kind;
                ptr->objectFile = defBase [defPtr]->subjectFile;
                ptr->objectLine = defBase [defPtr]->subjectLine;
                ptr->itemName = name;
                ptr->next = masterHead;
                masterHead = ptr;
            }
        }
        else
        {
            if (refBase [refPtr]->kind == IMPORT_LINK_FCN_TEMP)
                kind = IMPORT_EXTERN_FCN;
            else if (refBase [refPtr]->kind == IMPORT_LINK_VAR_TEMP)
                kind = IMPORT_EXTERN_VAR;
            else Cfx_AssertFailed (6);
            ptr = CreateNewItem (refBase [refPtr]->fileAttributed,
                                 refBase [refPtr]->subjectFile,
                                 refBase [refPtr]->subjectLine,
                                 NO_KIND, UNDEFINED_ITEM);
            ptr->kind = kind;
            ptr->objectFile = cfx_externalFileNumber;
            ptr->objectLine = -1;
            ptr->itemName = name;
            ptr->next = masterHead;
            masterHead = ptr;
        }
    }
} /* CorrelateLinkRefsAndDefs */


/************************************************************************/
/*      static void     CorrelateRefsDefsAndDecls (Item *listHead)      */
/*                                                                      */
/*   This function is used by CC1.  It goes through the Data LL and     */
/* pulls out and sorts all the                                          */
/************************************************************************/
static void     CorrelateRefsDefsAndDecls (listHead)
    Item        *listHead;
{
    int         counts [NO_KIND];
    int         numItems;
    Item        *ptr = listHead;
    Item        **refBase, **defBase, **declBase;
    int         ref = 0, def = 0, decl = 0, refPtr, defPtr, declPtr;
    char        *name;
    int         kind;

    (void) CountElementsInDataLL (listHead, counts);

    /* Allocate memory for the references */
    numItems = counts [IMPORT_FCN_REF_TEMP] + counts [IMPORT_VAR_REF_TEMP] + 1;
    refBase = (Item **) malloc (numItems * sizeof (Item *));
    if (refBase == NULL)
    {
        fprintf (stderr, "[17] No memory to allocate refBase\n");
        exit (117);
    }

    /* Allocate memory for the definitions */
    numItems = counts [EXPORT_PUBLIC_FCN_DEF] + counts [EXPORT_PUBLIC_VAR_DEF] +
           counts [EXPORT_STATIC_FCN_DEF] + counts [EXPORT_STATIC_VAR_DEF] + 1;
    defBase = (Item **) malloc (numItems * sizeof (Item *));
    if (defBase == NULL)
    {
        fprintf (stderr, "[18] No memory to allocate defBase\n");
        exit (118);
    }

    /* Allocate memory for the declarations */
    numItems = counts [EXPORT_PUBLIC_FCN_DECL] + 
               counts [EXPORT_PUBLIC_VAR_DECL] +
               counts [EXPORT_STATIC_FCN_DECL] + 1;
    declBase = (Item **) malloc (numItems * sizeof (Item *));
    if (declBase == NULL)
    {
        fprintf (stderr, "[19] No memory to allocate declBase\n");
        exit (119);
    }

    while (ptr != NULL)
    {
        switch (ptr->kind)
        {
            case IMPORT_FCN_REF_TEMP:
            case IMPORT_VAR_REF_TEMP:
                refBase [ref++] = ptr;
                break;
            case EXPORT_PUBLIC_FCN_DEF:
            case EXPORT_PUBLIC_VAR_DEF:
            case EXPORT_STATIC_FCN_DEF:
            case EXPORT_STATIC_VAR_DEF:
                defBase [def++] = ptr;
                break;
            case EXPORT_PUBLIC_FCN_DECL:
            case EXPORT_PUBLIC_VAR_DECL:
            case EXPORT_STATIC_FCN_DECL:
                declBase [decl++] = ptr;
                break;
            default:
                break;
        }
        ptr = ptr->next;
    }

    qsort (refBase, ref, sizeof (Item *), SortByName);
    qsort (defBase, def, sizeof (Item *), SortByName);
    qsort (declBase, decl, sizeof (Item *), SortByName);

    /* Allocate an element at the end with type NO_KIND */
    refBase [ref] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);
    refBase [ref]->itemName = "\377";
    defBase [def] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);
    defBase [def]->itemName = "\377";
    declBase [decl] = CreateNewItem (-1, -1, -1, NO_KIND, UNDEFINED_ITEM);
    declBase [decl]->itemName = "\377";

    /* First we add the IMPORT_IMPLEMENT.  For every definition, we     */
    /* emit an IMPORT_IMPLEMENT for every declaration in a different    */
    /* file.                                                            */
    declPtr = 0;
    for (defPtr = 0 ; defPtr < def ; defPtr++)
    {
        name = defBase [defPtr]->itemName;
        while (strcmp (name, declBase [declPtr]->itemName) > 0)
        {
            declPtr++;
        }
        while (strcmp (name, declBase [declPtr]->itemName) == 0)
        {
            
            if (defBase [defPtr]->fileAttributed != 
                declBase [declPtr]->fileAttributed)
            {
                
                if (defBase [defPtr]->kind == EXPORT_PUBLIC_FCN_DEF)
                {
                    kind = IMPORT_IMPLEMENT_PUBLIC_FCN;
                }
                else if (defBase [defPtr]->kind == EXPORT_PUBLIC_VAR_DEF)
                {
                    kind = IMPORT_IMPLEMENT_PUBLIC_VAR;
                }
                else if (defBase [defPtr]->kind == EXPORT_STATIC_FCN_DEF)
                {
                    kind = IMPORT_IMPLEMENT_STATIC_FCN;
                }
                else if (defBase [defPtr]->kind == EXPORT_STATIC_VAR_DEF)
                {
                    kind = IMPORT_IMPLEMENT_STATIC_VAR;
                }
                else Cfx_AssertFailed (3);
                ptr = CreateNewItem (defBase [defPtr]->fileAttributed,
                                     defBase [defPtr]->subjectFile,
                                     defBase [defPtr]->subjectLine,
                                     kind, UNDEFINED_ITEM);
                ptr->objectFile = declBase [declPtr]->subjectFile;
                ptr->objectLine = declBase [declPtr]->subjectLine;
                ptr->itemName = name;
            }
            declPtr++;
        }
    }

    /* Second we add the IMPORT_x_DEF or IMPORT_x_DECL and IMPORT_LINK  */
    /* if necessary                                                     */
    declPtr = 0;
    defPtr = 0;
    refPtr = 0;
    while (refPtr < ref)
    {
        name = refBase [refPtr]->itemName;

        /* First, try and match it with a definition */
        while (strcmp (name, defBase [defPtr]->itemName) > 0)
        {
            defPtr++;
        }
        if (strcmp (name, defBase [defPtr]->itemName) == 0)
        {
            if (refBase [refPtr]->fileAttributed != 
                defBase [defPtr]->fileAttributed)
            {
                
                if (defBase [defPtr]->kind == EXPORT_PUBLIC_FCN_DEF)
                {
                    kind = IMPORT_PUBLIC_FCN_DEF;
                }
                else if (defBase [defPtr]->kind == EXPORT_PUBLIC_VAR_DEF)
                {    
                    kind = IMPORT_PUBLIC_VAR_DEF;
                }
                else if (defBase [defPtr]->kind == EXPORT_STATIC_FCN_DEF)
                {    
                    kind = IMPORT_STATIC_FCN_DEF;
                }
                else if (defBase [defPtr]->kind == EXPORT_STATIC_VAR_DEF)
                {    
                    kind = IMPORT_STATIC_VAR_DEF;
                }
                else Cfx_AssertFailed (2);
                ptr = CreateNewItem (refBase [refPtr]->fileAttributed,
                                     refBase [refPtr]->subjectFile,
                                     refBase [refPtr]->subjectLine,
                                     kind, refBase [refPtr]->scopeName ); /* UNDEFINED_ITEM); */
                ptr->objectFile = defBase [defPtr]->subjectFile;
                ptr->objectLine = defBase [defPtr]->subjectLine;
                ptr->itemName = name;
            }
        }
        else
        {
            /* if there was no match on a definition, try a declaration */
            int origDeclPtr;

            while (strcmp (name, declBase [declPtr]->itemName) > 0)
            {
                declPtr++;
            }

            origDeclPtr = declPtr;

            while (strcmp (name, declBase [declPtr]->itemName) == 0)
            {
                if (refBase [refPtr]->fileAttributed != 
                    declBase [declPtr]->fileAttributed)
                {
                    /* First place the decl import for each decl */
                    if (declBase [declPtr]->kind == EXPORT_PUBLIC_FCN_DECL)
                        kind = IMPORT_PUBLIC_FCN_DECL;
                    else if (declBase [declPtr]->kind == EXPORT_PUBLIC_VAR_DECL)
                        kind = IMPORT_PUBLIC_VAR_DECL;
                    else Cfx_AssertFailed (1);
                    ptr = CreateNewItem (refBase [refPtr]->fileAttributed,
                                         refBase [refPtr]->subjectFile,
                                         refBase [refPtr]->subjectLine,
                                         kind, UNDEFINED_ITEM);
                    ptr->objectFile = declBase [declPtr]->subjectFile;
                    ptr->objectLine = declBase [declPtr]->subjectLine;
                    ptr->itemName = name;

                }
                declPtr++;
            }

            declPtr = origDeclPtr;

            /* Now put in the LINK import */
            if (refBase [refPtr]->kind == IMPORT_FCN_REF_TEMP)
                kind = IMPORT_LINK_FCN_TEMP;
            else if (refBase [refPtr]->kind == IMPORT_VAR_REF_TEMP)
                kind = IMPORT_LINK_VAR_TEMP;
            ptr = CreateNewItem(refBase [refPtr]->fileAttributed,
                                 refBase [refPtr]->subjectFile,
                                 refBase [refPtr]->subjectLine,
                                 kind, refBase [refPtr]->scopeName );
            ptr->itemName = name;
        }

        /* Incremement refPtr, check to make certain new one is not identical */
        /* to the previous one (this step eliminates duplicates               */
        refPtr++;
        while (ITEM_PTR_RECS_EQUAL (refBase [refPtr], refBase [refPtr - 1]))
        {
            refPtr++;
        }
    } /* while */
} /* CorrelateRefsDefsAndDecls */


/************************************************************************/
/*      static int      CountElementsInDataLL (Item *listHead,          */
/*                                             int counts [])           */
/************************************************************************/
static int      CountElementsInDataLL (listHead, counts)
    Item        *listHead;
    int         counts [];
{
    int         cnt, total = 0;
    Item        *ptr = listHead;

    for (cnt = 0 ; cnt < NO_KIND ; cnt++) counts [cnt] = 0;

    while (ptr != NULL)
    {
        total++;
        counts [ptr->kind]++;
        ptr = ptr->next;
    }

    return (total);
} /* CountElementsInDataLL */


/************************************************************************/
/*      static Item     *CreateNewItem (int fileAttributed, subjectFile,*/
/*                                      subjectLine, kind, scopeName)              */
/************************************************************************/
static Item     *CreateNewItem (fileAttributed, subjectFile, subjectLine, kind, scopeName)
    int fileAttributed, subjectFile, subjectLine, kind;
    char *scopeName;
{
    Item        *ptr; 

    ptr = (Item *) malloc (sizeof (Item));
    if (ptr == NULL)
    {
        fprintf (stderr, "[20] No memory for item.\n");
        exit (120);
    }

    ptr->kind = kind;
    ptr->fileAttributed = fileAttributed;
    ptr->subjectFile = subjectFile;
    ptr->subjectLine = subjectLine;
    ptr->objectFile = -1;
    ptr->objectLine = -1;
    ptr->itemName = UNDEFINED_ITEM;
    ptr->scopeName = scopeName;

    if (kind == NO_KIND)
    {
        /* Not a real item, don't insert it in the list */
        return (ptr);
    }

    ptr->next = mainHead;
    mainHead = ptr;

    return (ptr);
} /* CreateNewItem */


/************************************************************************/
/*      static void     FileReadError (char *line)                      */
/************************************************************************/
static void     FileReadError (line)
    char *line;
{
    if (line == NULL)
    {
        fprintf (stderr, "[21a] Error reading file\n");
    }
    else
    {
        fprintf (stderr, "[21b] Error while reading line: \"%s\"\n", line);
    }
    exit (121);
} /* FileReadError */


/************************************************************************/
/*      static int      GetFileNameForRead ()                           */
/*                                                                      */
/*   This routine tries to find a filename suitable for opening for     */
/* write.  It returns a pointer to the file name.  It tries filenames   */
/* in the following order:                                              */      
/*                                                                      */
/*   firstName = <1Dir>/<1File> and secondName = <2Dir>/<2File>         */
/* 1st choice for name is <2Dir>/.cfx.<prefix>.<2File>                  */
/* 2nd choice for name is <1Dir>/.cfx.<prefix>.<1File>                  */
/* 3rd choice for name is .cfx.<prefix>.<1File>                         */
/* 4th choice for name is .cfx.<prefix>                                 */
/************************************************************************/
static char     *GetFileNameForRead (firstName, secondName, prefix)
    char *firstName, *secondName;
    int prefix;
{
    char        *fileName;

    /* At this point, if secondName is NULL, "" or "-", then we are not */
    /* not going to use <2Dir>/.cfx.<prefix>.<2File>.  Otherwise, we    */
    /* *must* use it or fail immediately.                               */
    if ((secondName != NULL) && (strcmp (secondName, "") != 0) &&
        (strcmp (secondName, "-") != 0))
    {
        fileName = (char *) malloc (strlen (secondName) + 30);
        Cfx_CreateFileName (secondName, fileName, -1);
        if (CheckReadable (fileName))
        {
            return (fileName);
        }
        fprintf (stderr, "[22] Unable to open \"%s\" for read.\n", fileName);
        exit (122);
    }

    /* At this point, if firstName is NULL or "", then we are not       */
    /* not going to use choice 2.                                       */
    /* A choice is successful if one can open the file for put and then */
    /* unlink it.                                                       */
    if ((firstName != NULL) && (strcmp (firstName, "") != 0) &&
        (strcmp (firstName, "-") != 0))
    {
        char    *ptr;
        /****************/
        /* Try choice 2 */
        /****************/
        fileName = (char *) malloc (strlen (firstName) + 30);
        Cfx_CreateFileName (firstName, fileName, prefix);
        if (CheckReadable (fileName))
        {
            return (fileName);
        }

        /****************/
        /* Try choice 3 */
        /****************/
        if (strrchr (firstName, '/') == NULL)
        {       
            ptr = firstName;
        }
        else
        {
            ptr = strrchr (firstName, '/') + 1;
        }
        Cfx_CreateFileName (ptr, fileName, prefix);
        if (CheckReadable (fileName))
        {
            return (fileName);
        }
        fprintf (stderr, "[23] Unable to open \"%s\" for write.\n", fileName);
        exit (123);
    }

    /* Use the default path name */
    if (prefix == CPP)
    {
        return (CFX_DEFAULT_CPP_FILE);
    }
    else if (prefix == CC1)
    {
        return (CFX_DEFAULT_CC1_FILE);
    }

    fprintf (stderr, "[24] Unrecognized prefix in GetFileNameForRead\n");
    exit (124);
} /* GetFileNameForRead */


/************************************************************************/
/*      static int      GetFileNameForWrite ()                          */
/*                                                                      */
/*   This routine tries to find a filename suitable for opening for     */
/* write.  It returns a pointer to the file name.  It tries filenames   */
/* in the following order:                                              */      
/*                                                                      */
/*   firstName = <1Dir>/<1File> and secondName = <2Dir>/<2File>         */
/* 1st choice for name is <2Dir>/.cfx.<prefix>.<2File>                  */
/* 2nd choice for name is <1Dir>/.cfx.<prefix>.<1File>                  */
/* 3rd choice for name is .cfx.<prefix>.<1File>                         */
/* 4th choice for name is .cfx.<prefix>                                 */
/************************************************************************/
static char     *GetFileNameForWrite (firstName, secondName, prefix)
    char *firstName, *secondName;
    int prefix;
{
    char        *fileName;

    /* At this point, if secondName is NULL, "" or "-", then we are not */
    /* not going to use <2Dir>/.cfx.<prefix>.<2File>.  Otherwise, we    */
    /* *must* use it or fail immediately.                               */
    if ((secondName != NULL) && (strcmp (secondName, "") != 0) &&
        (strcmp (secondName, "-") != 0))
    {
        /****************/
        /* Try choice 1 */
        /****************/
        fileName = (char *) malloc (strlen (secondName) + 30);
        Cfx_CreateFileName (secondName, fileName, -1);
        if (CheckWriteable (fileName))
        {
            return (fileName);
        }
        fprintf (stderr, "[25] Unable to open \"%s\" for write.\n", fileName);
        exit (125);
    }

    /* At this point, if firstName is NULL, "" or "-", then we are not  */
    /* not going to use choice 2 or 3.  Otherwise, one of choice 2 or 3 */
    /* *must* succeed or fail immediately.                              */
    if ((firstName != NULL) && (strcmp (firstName, "") != 0) &&
        (strcmp (firstName, "-") != 0))
    {
        char    *ptr;

        /****************/
        /* Try choice 2 */
        /****************/
        fileName = (char *) malloc (strlen (firstName) + 30);
        Cfx_CreateFileName (firstName, fileName, prefix);
        if (CheckWriteable (fileName))
        {
            return (fileName);
        }

        /****************/
        /* Try choice 3 */
        /****************/
        if (strrchr (firstName, '/') == NULL)
        {       
            ptr = firstName;
        }
        else
        {
            ptr = strrchr (firstName, '/') + 1;
        }
        Cfx_CreateFileName (ptr, fileName, prefix);
        if (CheckWriteable (fileName))
        {
            return (fileName);
        }
        fprintf (stderr, "[26] Unable to open \"%s\" for write.\n", fileName);
        exit (126);
    }

    /* Use the default path name */
    if (prefix == CPP)
    {
        return (CFX_DEFAULT_CPP_FILE);
    }
    else if (prefix == CC1)
    {
        return (CFX_DEFAULT_CC1_FILE);
    }

    fprintf (stderr, "[27] Unrecognized prefix in GetFileNameForWrite\n");
    exit (127);
} /* GetFileNameForWrite */


/************************************************************************/
/*      static void     InitializeFile (char *fileName)                 */
/************************************************************************/
static void     InitializeFile (fileName)
    char *fileName;
{
    FILE        *f;
   
    unlink (fileName);
} /* InitializeFile */


/************************************************************************/
/*      static void     MallocString (char **itemToMalloc, char *str)   */
/************************************************************************/
static void     MallocString (itemToMalloc, str)
    char **itemToMalloc, *str;
{
    *itemToMalloc = (char *) malloc (strlen (str) + 1);
    if (*itemToMalloc == NULL)
    {
        fprintf (stderr, "[28] No memory for string \"%s\".\n", str);
        exit (128);
    }
    strcpy (*itemToMalloc, str);
} /* MallocString */


/************************************************************************/
/*      static void     MallocStringLen (char **itemToMalloc,           */
/*                                       char *str, int len)            */
/************************************************************************/
static void     MallocStringLen (itemToMalloc, str, len)
    char        **itemToMalloc, *str;
    int         len;
{
    *itemToMalloc = (char *) malloc (len + 1);
    if (*itemToMalloc == NULL)
    {
        fprintf (stderr, 
                 "[29] No memory to allocate space for string \"%s\".\n", 
                 str);
        exit (129);
    }
    strncpy (*itemToMalloc, str, len);
    (*itemToMalloc) [len] = 0;
} /* MallocString */


/************************************************************************/
/*      static void     ReadAndVerifyHeader (FILE *f, char *header)     */
/************************************************************************/
static void     ReadAndVerifyHeader (f, fileName, header)
    FILE *f;
    char *fileName, *header;
{
    char        line [1024];

    if (fgets (line, 1024, f) == NULL) FileReadError (NULL);
    if (strncmp (line, header, strlen (header) - 6))
    {
        fprintf (stderr, "[30] %s did not have expected header!\n", fileName);
        fprintf (stderr, "Actually read: \"%s\"\n", line);
        fprintf (stderr, "Should be:     \"%s\"\n", header);
        exit (130);
    }
} /* ReadAndVerifyHeader */



static char *Fgets(char *line, int mxl, FILE *fp)
{
    int len;

    if (fgets(line, mxl, fp) == NULL)
	return NULL;

    len = strlen(line);

    if (line[len-1] == '\n')
	line[--len] = '\0';

    if (line[len-1] == '\r')
	line[--len] = '\0';

    return line;
}

    


/************************************************************************/
/*      static void     ReadFileListIntoTable (FILE *f)                 */
/************************************************************************/
static void     ReadFileListIntoTable (f)
    FILE *f;
{
    char        line [1024];
    int         i;

    /**********************/
    /* Read past comments */
    /**********************/
    SkipComments (f);

    /************************/
    /* Read file table size */
    /************************/
    if (fgets (line, 1024, f) == NULL) FileReadError (NULL);
    if (sscanf (line, "%d", &numFiles) != 1) FileReadError (line);

    /*******************/
    /* Read file table */
    /*******************/
    fileTable = (Files *) malloc ((numFiles + 1) * sizeof (Files));
    if (fileTable == NULL)
    {
        fprintf (stderr, "[31] No memory for file table.\n");
        exit (131);
    }
    for (i = 1 ; i <= numFiles ; i++)
    {
        if (Fgets (line, 1024, f) == NULL) FileReadError (NULL);
        MallocString (&(fileTable [i].fileName), line);

        if (Fgets (line, 1024, f) == NULL) FileReadError (NULL);
        MallocString (&(fileTable [i].shortFileName), line);
    }

    /************************/
    /* Set up the zero item */
    /************************/
    fileTable [0].fileName = "(No File)";
    fileTable [0].shortFileName = "(No File)";
    fileTable [0].comparisonValue = 0;
    fileTable [0].mainFile = 0;
    fileTable [0].interfaceFileWritten = 0;
} /* ReadFileListIntoTable */


/************************************************************************/
/*      static void     ReadMRDataIntoDataLL (FILE *f)                  */
/************************************************************************/
static void     ReadMRDataIntoDataLL (f)
    FILE *f;
{
    char        line [1024];
    int         i, numElements;
    int         kind, fileAttributed, subjectFile, subjectLine;
    int         objectFile, objectLine;
    char        itemName [1024];

    /**********************/
    /* Read past comments */
    /**********************/
    SkipComments (f);

    /************************/
    /* Read main file size */
    /************************/
    if (fgets (line, 1024, f) == NULL) FileReadError (NULL);
    if (sscanf (line, "%d", &numElements) != 1) FileReadError (line);

    /******************/
    /* Read main file */
    /******************/
    for (i = 0 ; i < numElements ; i++)
    {
        Item    *newItem;

        if (fgets (line, 1024, f) == NULL) FileReadError (NULL);
        if (sscanf (line, " %4d %4d %4d %6d %4d %6d %s\n", &kind, 
                    &fileAttributed, &subjectFile, &subjectLine, &objectFile, 
                    &objectLine, itemName) != 7)
        {
            FileReadError (line);
        }
        newItem = CreateNewItem (fileAttributed, subjectFile, subjectLine, 
                                 kind, UNDEFINED_ITEM);
        newItem->objectFile = objectFile;
        newItem->objectLine = objectLine;
        MallocString (&(newItem->itemName), itemName);
    }
} /* ReadMRDataIntoDataLL */


/************************************************************************/
/*      static void     SkipAn (base, ptr, type)                        */
/************************************************************************/
static void     SkipAn (base, ptr, type, initialFile)
    Item        **base;
    int         *ptr, type, initialFile;
{
    while ((base [*ptr]->kind == type) &&
           (base [*ptr]->fileAttributed == initialFile))
    {
        (*ptr)++;
    }
} /* SkipAn */


/************************************************************************/
/*      static void     SkipComments (f)                                */
/************************************************************************/
static void     SkipComments (f)
    FILE *f;
{
    int         pos;
    char        line [1026];
    
    do
    {
        pos = ftell (f);
        if (fgets (line, 1024, f) == NULL)
        {
            FileReadError (NULL);
        }
    }
    while (line [0] == '#');
    fseek (f, pos, 0);
} /* SkipComments */


/************************************************************************/
/*      static void     Cfx_AssertFailed (error)                        */
/************************************************************************/
static void     Cfx_AssertFailed (error)
    int error;
{
    fprintf (stderr, "Assertion Failure #%d\n", error);
    exit (200 + error);
} /* Cfx_AssertFailed */


/************************************************************************/
/*      static int SortByName (first, second)                           */
/************************************************************************/
static int      SortByName (first, second)
    Item        **first, **second;
{
    return (strcmp ((*first)->itemName, (*second)->itemName));
} /* SortByName */


/************************************************************************/
/*      static char     *TimeString (void)                              */
/************************************************************************/
static char     *TimeString ()
{
  time_t        t;
  static char   timeStr [50];

  t = time (NULL);
  strcpy (timeStr, ctime (&t));
  timeStr [strlen (timeStr) - 1] = 0;
  return (timeStr);
} /* TimeString */


/************************************************************************/
/*      static void     WriteHRExport (f, base, ptr, type, title)       */
/************************************************************************/
static void     WriteHRExport (f, base, ptr, type, initialFile, title)
    FILE        *f;
    Item        **base;
    int         *ptr, type, initialFile;
    char        *title;
{
    int         start = 1, col = 80;
    char        *last = "";

    if ((base [*ptr]->kind == type) &&
        (base [*ptr]->fileAttributed == initialFile))
    {
        if (!noHrOption)
        {
            fprintf (f, "        %s:", title);
        }
        while ((base [*ptr]->kind == type) && 
               (base [*ptr]->fileAttributed == initialFile))
        {
            if (!noHrOption && (strcmp (last, base [*ptr]->itemName) != 0))
            {
                if (col + strlen (base [*ptr]->itemName) > 78)
                {
                    fprintf (f, "\n            %s", base [*ptr]->itemName);
                    col = 12 + strlen (base [*ptr]->itemName);
                }
                else
                {
                    fprintf (f, ", %s", base [*ptr]->itemName);
                    col += strlen (base [*ptr]->itemName) + 2;
                }
                last = base [*ptr]->itemName;
            }
            (*ptr)++;
        }
        if (!noHrOption)
        {
            fprintf (f, "\n");
        }
    }
} /* WriteHRExport */


/************************************************************************/
/*      static void     WriteHRImport (f, m, base, ptr, type, title)    */
/************************************************************************/
static void     WriteHRImport (f, m, base, ptr, type, objectFile, initialFile,
                               title)
    FILE        *f, *m;
    Item        **base;
    int         *ptr, type, objectFile, initialFile;
    char        *title;
{
    int         start = 1, col = 80;
    char        *last = "";

    if ((base [*ptr]->kind == type) &&
        (base [*ptr]->objectFile == objectFile) &&
        (base [*ptr]->fileAttributed == initialFile))
    {
        if (!noHrOption)
        {
            fprintf (f, "            %s:", title);
        }
        if (!noMinOption)
        {
            fprintf (m, "%d ", type);
        }
        while (start || 
               ((base [*ptr]->kind == type) &&
                (base [*ptr]->fileAttributed == initialFile) &&
                (base [*ptr]->objectFile == base [*ptr - 1]->objectFile)))
        {
            start = 0;
            if (!noHrOption && (strcmp (last, base [*ptr]->itemName) != 0))
            {
                if (col + strlen (base [*ptr]->itemName) > 78)
                {
                    fprintf (f, "\n                %s", base [*ptr]->itemName);
                    col = 16 + strlen (base [*ptr]->itemName);
                }
                else
                {
                    fprintf (f, ", %s", base [*ptr]->itemName);
                    col += strlen (base [*ptr]->itemName) + 2;
                }
                last = base [*ptr]->itemName;
            }
            (*ptr)++;
        }
        if (!noHrOption)
        {
            fprintf (f, "\n");
        }
    }
} /* WriteHRImport */


/************************************************************************/
/*      static void     WriteFileListFromFileLL (f)                     */
/************************************************************************/
static void     WriteFileListFromFileLL (f)
    FILE        *f;
{
    File        *ptr = fileHead;
    int         numFiles = 0;

    fprintf (f, "#\n");
    fprintf (f, "# The File Table\n");

    while (ptr != NULL)
    {
        numFiles++;
        ptr = ptr->next;
    }

    fprintf (f, "%d\n", numFiles);
    ptr = fileHead;
    while (ptr != NULL)
    {
        fprintf (f, "%s\n", ptr->fileName);
        fprintf (f, "%s\n", ptr->shortFileName);
        ptr = ptr->next;
    }
} /* WriteFileListFromFileLL */


/************************************************************************/
/*      static void     WriteFileListFromFileTable (FILE *f)            */
/************************************************************************/
static void     WriteFileListFromFileTable (f)
    FILE        *f;
{
    File        *ptr = fileHead;
    int         i;

    fprintf (f, "#\n");
    fprintf (f, "# The File Table\n");

    fprintf (f, "%d\n", numFiles);
    for (i = 1 ; i <= numFiles ; i++)
    {
        fprintf (f, "%s\n", fileTable [i].fileName);
        fprintf (f, "%s\n", fileTable [i].shortFileName);
    }
} /* WriteFileListFromFileTable */


/************************************************************************/
/*      static int      WriteHRDataFromDataTable (FILE *f, *m,          */
/*                                                Item **base,          */
/*                                                int ptr)              */
/************************************************************************/
static int      WriteHRDataFromDataTable (f, m, base, loc)
    FILE        *f, *m;
    Item        **base;
    int         loc;
{
    int         ptr;
    int         file = base [loc]->fileAttributed;
    int         cnt;
    int         fileIncludedCount;

    if (!noHrOption)
    {
        /* Write File Name in file */
        fprintf (f, "File: %s\n", fileTable [file].shortFileName);
    }

    if (!noMinOption)
    {
        /* Write File Name in Min File */
        fprintf (m, "%s\n", fileTable [file].shortFileName);
    }

    /**************************************/
    /* Write Human Readable Form of Table */
    /**************************************/
    ptr = loc;
    /********************************************************************/
    /* If the file table indicates that this file was only a main file  */
    /* (meaning that it was never included in another file), then we    */
    /* only indicate the things that it exports to the linker, the      */
    /* public variables and functions.  If this file was included in    */
    /* another file, then we ouput everything it exports                */
    /********************************************************************/
    if (fileTable [file].mainFile)
    {
        /****************************************************************/
        /* Note that main files only export link variables and fcns and */
        /* skip over everything else.                                   */
        /****************************************************************/
        if (base [ptr]->kind > EXPORT_PUBLIC_VAR_DEF)
        {
            if (!noHrOption)
            {
                fprintf (f, "    No exports\n");
            }
        }
        else
        {
            if (!noHrOption)
            {
                fprintf (f, "    Exports:\n");
            }
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_FCN_DEF, file, 
                           "Functions Exported");
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_VAR_DEF, file, 
                           "Variables Exported");
        }
        for (cnt = EXPORT_STATIC_FCN_DEF ; cnt < IMPORT_EXPORT_DIVIDER ; cnt++)
        {
            SkipAn (base, &ptr, cnt, file);
        }
    }
    else
    {
        /****************************************************************/
        /* Note that non-main files export everything                   */
        /****************************************************************/
        if (base [ptr]->kind >= IMPORT_EXPORT_DIVIDER)
        {
            if (!noHrOption)
            {
                fprintf (f, "    No exports\n");
            }
        }
        else
        {
            if (!noHrOption)
            {
                fprintf (f, "    Exports:\n");
            }
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_FCN_DEF, file, 
                           "Functions Exported");
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_VAR_DEF, file, 
                           "Variables Exported");
            WriteHRExport (f, base, &ptr, EXPORT_STATIC_FCN_DEF, file, 
                           "Static Functions Exported");
            WriteHRExport (f, base, &ptr, EXPORT_STATIC_VAR_DEF, file, 
                           "Static Variables Exported");
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_FCN_DECL, file, 
                           "Function Declarations Exported");
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_VAR_DECL, file, 
                           "Variable Declarations Exported");
            WriteHRExport (f, base, &ptr, EXPORT_STATIC_FCN_DECL, file, 
                           "Static Function Declarations Exported");
            WriteHRExport (f, base, &ptr, EXPORT_MACRO, file, 
                           "Macros Exported");
            WriteHRExport (f, base, &ptr, EXPORT_TYPE, file, 
                           "Types Exported");
            WriteHRExport (f, base, &ptr, EXPORT_ENUM_TAG, file, 
                           "Enumerated Type Tags Exported");
            WriteHRExport (f, base, &ptr, EXPORT_ENUM_ITEM, file, 
                           "Enumerated Types Items Exported");
            WriteHRExport (f, base, &ptr, EXPORT_STRUCT, file, 
                           "Structures Exported");
            WriteHRExport (f, base, &ptr, EXPORT_UNION, file, 
                           "Union Types Exported");
            WriteHRExport (f, base, &ptr, EXPORT_PUBLIC_FCN_DEF_END, file,/*SS*/
                           "Function Ends Exported");                     /*SS*/
            WriteHRExport (f, base, &ptr, EXPORT_STATIC_FCN_DEF_END, file,/*SS*/
                           "Static Function Ends Exported");              /*SS*/
            WriteHRExport (f, base, &ptr, EXPORT_ENUM_TAG_END, file,      /*SS*/
                           "Enumerated Type Tag End Exported");           /*SS*/
            WriteHRExport (f, base, &ptr, EXPORT_STRUCT_END, file,        /*SS*/
                           "Structure Ends Exported");                    /*SS*/
            WriteHRExport (f, base, &ptr, EXPORT_UNION_END, file,         /*SS*/
                           "Union Type Ends Exported");                   /*SS*/

        }
    }

    /********************************************************************/
    /* Now we do the imports.  First make certain that there are some   */
    /* imports.  If the ptr is now pointing to a record not associated  */
    /* with this file, then there are no imports for this file          */
    /********************************************************************/
    if (base [ptr]->fileAttributed != file)
    {
        if (!noHrOption)
        {
            fprintf (f, "    No imports\n");
        }
    }
    else
    {
        /****************************************************************/
        /* There are imports for this file.  We are going to want to    */
        /* a header (1) for every file that we import from and then     */
        /* for every file a (2) for every type of import that we do.    */
        /****************************************************************/
        int     start = 1;

        if (!noHrOption)
        {
            fprintf (f, "    Imports:\n");
        }

        /****************************************************************/
        /* When the ptr no longer points at this file, then we're done  */
        /* We should be going through this loop for every file that we  */
        /* import from.  i.e. one loop should catch all the imports     */
        /* from a given file.                                           */
        /****************************************************************/
        while (base [ptr]->fileAttributed == file)
        {
#if TW
            if (start || 
                (base [ptr]->objectFile != base [ptr - 1]->objectFile))
            {
#endif
                int objectFile = base [ptr]->objectFile;

                start = 0;
                if (!noHrOption)
                {
                    if (strcmp (
                             fileTable [objectFile].shortFileName,
                             fileTable [objectFile].fileName) == 0)
                    {
                        fprintf (f, "        %s:\n", 
                             fileTable [objectFile].fileName);
                    }
                    else
                    {
                        fprintf (f, "        %s (%s):\n", 
                             fileTable [objectFile].shortFileName,
                             fileTable [objectFile].fileName);
                    }
                }

                if (!noMinOption)
                {
                    fprintf (m, "%s ", fileTable [objectFile].shortFileName);
                }
        
                fileIncludedCount = 0;

                while ((base [ptr]->kind == INCLUDE_FILE) &&
                       (base [ptr]->objectFile == objectFile) &&
                       (base [ptr]->fileAttributed == file))
                {
                    fileIncludedCount++;
                    ptr++;
                    while (ITEM_PTR_RECS_EQUAL (base [ptr], base [ptr - 1]))
                    {
                        ptr++;
                    }
                }

                if (fileIncludedCount > 0)
                {
                    if (!noHrOption)
                    {
                        fprintf (f, "            File Included");
                        if (fileIncludedCount > 1)
                        {
                            fprintf (f, " (%d times)", fileIncludedCount);
                        }
                        fprintf (f, "\n");
                    }
                    if (!noMinOption)
                    {
                        fprintf (m, "%d ", INCLUDE_FILE);
                    }
                }

                WriteHRImport (f, m, base, &ptr, IMPORT_IMPLEMENT_PUBLIC_FCN,
                           objectFile, file, "Functions Implemented");
                WriteHRImport (f, m, base, &ptr, IMPORT_IMPLEMENT_PUBLIC_VAR,
                           objectFile, file, "Variables Implemented");
                WriteHRImport (f, m, base, &ptr, IMPORT_IMPLEMENT_STATIC_FCN,
                           objectFile, file, "Static Functions Implemented");
                WriteHRImport (f, m, base, &ptr, IMPORT_IMPLEMENT_STATIC_VAR,
                           objectFile, file, "Static Variables Implemented");
                WriteHRImport (f, m, base, &ptr, IMPORT_PUBLIC_FCN_DEF, 
                           objectFile, file, "Function Definition Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_PUBLIC_VAR_DEF, 
                           objectFile, file, "Variables Definition Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_PUBLIC_FCN_DECL, 
                           objectFile, file, "Function Declarations Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_PUBLIC_VAR_DECL, 
                           objectFile, file, "Variables Declarations Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_STATIC_FCN_DEF, 
                           objectFile, file, "Static Functions Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_STATIC_VAR_DEF, 
                           objectFile, file, "Static Variables Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_LINK_FCN, 
                           objectFile, file, "Link Time Functions Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_LINK_VAR, 
                           objectFile, file, "Link Time Variables Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_EXTERN_FCN, 
                           objectFile, file, "External Functions Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_EXTERN_VAR, 
                           objectFile, file, "External Variables Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_CODE_AND_VARS, 
                           objectFile, file, "Code and Variables Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_MACRO,
                           objectFile, file, "Macros Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_TYPE,
                           objectFile, file, "Types Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_ENUM_TAG,
                           objectFile, file, "Enumerated Type Tags Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_ENUM_ITEM,
                           objectFile, file, "Enumerated Types Items Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_STRUCT,
                           objectFile, file, "Structures Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_UNION,
                           objectFile, file, "Union Types Imported");
                WriteHRImport (f, m, base, &ptr, IMPORT_FCN_REF_TEMP,
                           objectFile, file, "Function Reference (Temp)");
                WriteHRImport (f, m, base, &ptr, IMPORT_VAR_REF_TEMP,
                           objectFile, file, "Variables Reference (Temp)");
                WriteHRImport (f, m, base, &ptr, IMPORT_LINK_FCN_TEMP,
                           objectFile, file, "Function Reference Link (Temp)");
                WriteHRImport (f, m, base, &ptr, IMPORT_LINK_VAR_TEMP, 
                           objectFile, file, "Variable Reference Link (Temp)");
                if (!noMinOption)
                {
                    fprintf (m, "\n");
                }
#if TW
            }
#endif
        }
    }

    if (!noMinOption)
    {
        fprintf (m, "---\n");
    }
    return (ptr);
} /* WriteHRDataFromDataTable */


/************************************************************************/
/*      static void     WriteMRDataFromDataLL_gerry (FILE *f,                 */
/*                                             Item *listHead)          */
/************************************************************************/
static void     WriteMRDataFromDataLL_gerry (f, listHead)
    FILE        *f;
    Item        *listHead;
{
    Item        *ptr = listHead;
    int         numElements = 0;

    while (ptr != NULL)
    {
        if ((ptr->kind != IMPORT_FCN_REF_TEMP) &&
            (ptr->kind != IMPORT_VAR_REF_TEMP))
        {
            numElements++;
        }
        ptr = ptr->next;
    }

    fprintf (f, "#\n");
    fprintf (f, "# Imports/Exports\n");
    fprintf (f, "#\n");
    fprintf (f, "#     File   Subject     Object\n");
    fprintf (f, "#Kind Attr File  Line  File  Line          Scope Name         Item Name\n");
    fprintf (f, "#---- ---- ---- ------ ---- ------ -------------------------- ---------\n");

    fprintf (f, "%d\n", numElements);

    ptr = listHead;
    while (ptr != NULL)
    {
        if ((ptr->kind != IMPORT_FCN_REF_TEMP) &&
            (ptr->kind != IMPORT_VAR_REF_TEMP))
        {
            if ( strcmp(ptr ->scopeName, UNDEFINED_ITEM) == 0) ptr -> scopeName = "-";
            fprintf (f, " %4d %4d %4d %6d %4d %6d %25s %s\n", ptr->kind, 
                     ptr->fileAttributed, ptr->subjectFile, ptr->subjectLine,
                     ptr->objectFile, ptr->objectLine,
                     ptr->scopeName, ptr->itemName);
        }
        ptr = ptr->next;
    }
} /* WriteMRDataFromDataLL_gerry */


/************************************************************************/
/*      static void     WriteMRDataFromDataLL (FILE *f,                 */
/*                                             Item *listHead)          */
/************************************************************************/
static void     WriteMRDataFromDataLL (f, listHead)
    FILE        *f;
    Item        *listHead;
{
    Item        *ptr = listHead;
    int         numElements = 0;

    while (ptr != NULL)
    {
        if ((ptr->kind != IMPORT_FCN_REF_TEMP) &&
            (ptr->kind != IMPORT_VAR_REF_TEMP))
        {
            numElements++;
        }
        ptr = ptr->next;
    }

    fprintf (f, "#\n");
    fprintf (f, "# Imports/Exports\n");
    fprintf (f, "#\n");
    fprintf (f, "#     File   Subject     Object\n");
    fprintf (f, "#Kind Attr File  Line  File  Line  Item Name\n");
    fprintf (f, "#---- ---- ---- ------ ---- ------ ---------\n");

    fprintf (f, "%d\n", numElements);

    ptr = listHead;
    while (ptr != NULL)
    {
        if ((ptr->kind != IMPORT_FCN_REF_TEMP) &&
            (ptr->kind != IMPORT_VAR_REF_TEMP))
        {
            fprintf (f, " %4d %4d %4d %6d %4d %6d %s\n", ptr->kind, 
                     ptr->fileAttributed, ptr->subjectFile, ptr->subjectLine,
                     ptr->objectFile, ptr->objectLine,
                     ptr->itemName);
        }
        ptr = ptr->next;
    }
} /* WriteMRDataFromDataLL */


/************************************************************************/
/*      static int      WriteMRDataFromDataTable (FILE *f, Item **base, */
/*                                                int loc, numElements) */
/************************************************************************/
static int      WriteMRDataFromDataTable (f, base, loc)
    FILE        *f;
    Item        **base;
    int         loc;
{
    int         i, match;
    Item        *ptr;
    int         initialFile = base [loc]->fileAttributed;

    /* Print Header */
    fprintf (f, "#\n");
    fprintf (f, "# Imports/Exports from %s\n", 
             fileTable [initialFile].shortFileName);
    fprintf (f, "#\n");
    fprintf (f, "#     File   Subject     Object\n");
    fprintf (f, "#Kind Attr File  Line  File  Line  Item Name\n");
    fprintf (f, "#---- ---- ---- ------ ---- ------ ---------\n");

    /* Print File Names */
    fprintf (f, "%s %s\n", fileTable [initialFile].shortFileName,
             fileTable [initialFile].fileName);

    /* Print Number of Elements */
    i = loc;
    while (base [i]->fileAttributed == initialFile)
    {
        i++;
    }
    fprintf (f, "%d\n", i - loc);

    i = loc;
    while (base [i]->fileAttributed == initialFile)
    {
        match = 0;
        ptr = base [i];
        if (i != loc)
        {
            if (ITEM_PTR_RECS_EQUAL (ptr, base [i - 1]))
            {
                match = 1;
            }
        }
        if (!match)
        {
            fprintf (f, " %4d %4d %4d %6d %4d %6d %s\n", ptr->kind, 
                     ptr->fileAttributed, ptr->subjectFile, ptr->subjectLine,
                     ptr->objectFile, ptr->objectLine,
                     ptr->itemName);
        }
        i++;
    }
    return (i);
} /* WriteMRDataFromDataTable */


